#include <cppcms/service.h>
#include <cppcms/application.h>
#include <cppcms/applications_pool.h>
#include <cppcms/cache_interface.h>
#include <cppcms/http_request.h>
#include <cppcms/http_response.h>
#include <dbixx/dbixx.h>
#include <iostream>
#include <sstream>
#include "data.h"

#if defined(MYMKD)
namespace {
    using std::string;
    class Markdown {
    public:
        enum State {
            BLANK,
            PARA,
            LIST,
            QUOTE
        };

        State state ;

        std::string outStream;

        std::string const &input;

        void setState(State s)
        {
            if(s==state)
                return;
            switch(state) {
            case PARA:
                outStream.append("</p>\n");
                break;
            case LIST:
                outStream.append("</ul>\n");
                break;
            case QUOTE:
                outStream.append("</blockquote>\n");
                break;
            default:
                ;
            }

            state = s;

            switch(state) {
            case PARA:
                outStream.append("<p>");
                break;
            case LIST:
                outStream.append("<ul>\n");
                break;
            case QUOTE:
                outStream.append("<blockquote>\n");
                break;
            default:
                ;
            }
        }

        Markdown(std::string const &tmp) : 
            input(tmp)
        {
        }

        string convert()
        {
            size_t len = input.size();
            outStream.reserve(len*11/10);
            size_t pos = 0;
            state = BLANK;
            while(pos < len) {
                size_t line_end = input.find('\n',pos);
                if(line_end == std::string::npos)
                    line_end = len;
                else
                    line_end++;
                if(input[pos]=='#') {
                    setState(BLANK);
                    outStream.append("<h1>");
                    processText(pos+1,line_end);
                    outStream.append("</h1>\n");
                }
                else if(input[pos]=='-') {
                    setState(LIST);
                    outStream.append("<li>");
                    processText(pos+1,line_end);
                    outStream.append("</li>\n");
                }
                else if(input[pos]=='>') {
                    setState(QUOTE);
                    processText(pos+1,line_end);
                    outStream.append("\n");
                }
                else {
                    bool is_blank = true;
                    for(unsigned i=pos;is_blank && i<line_end;i++) {
                        char c = input[i];
                        switch(c) {
                        case ' ':
                        case '\t':
                        case '\n':
                        case '\r':
                            break;
                        default:
                            is_blank = false;
                            break;
                        }
                    }
                    if(is_blank)
                        setState(BLANK);
                    else { 
                        setState(PARA);
                        processText(pos,line_end);
                        outStream.append("\n");
                    }
                }
                pos = line_end;
            }
            setState(BLANK);
            std::string res;
            res.swap(outStream);
            return res;
        }

        void processText(int start,int end)
        {
            while(start < end && (input[start]==' ' || input[start]=='\t'))
                start++;
            for(int i=start;i<end;i++) {
                char c=input[i];
                switch(c) {
                case '<':
                    outStream.append("&lt;");
                    break;
                case '>':
                    outStream.append("&gt;");
                    break;
                case '&':
                    outStream.append("&amp;");
                    break;
                case '\r':
                case '\n':
                    break;
                default:
                    outStream+=(c);
                    break;
                }
            }
        }

    };

} // namespace

std::string to_markdown(std::string const &s)
{
		Markdown m(s);
		return m.convert();
}

#else // discount

extern "C" {
	#include <mkdio.h>
}

std::string to_markdown(std::string const &s)
{
	/// It is safe to const cast as mkd_string does not 
	/// alter original string
    char const *str=s.c_str();
    int len = s.size();
    int flags =0;
	MMIOT *doc = mkd_string(const_cast<char *>(str),len,flags);
	if(!doc) {
		throw std::runtime_error("Failed to read document");
	}

	mkd_compile(doc,flags);
	
	std::string result;
	result.reserve(len);
	char *content_ptr = 0;
	int content_size = 0;

	content_size = mkd_document(doc,&content_ptr);
	result.append(content_ptr,content_size);
	mkd_cleanup(doc);

	return result;
}


#endif

class test: public cppcms::application {
public:
    test(cppcms::service &s) :
        cppcms::application(s) 
    {
        sql.driver("mysql");
        sql.param("dbname","cppcms");
        sql.param("username","root");
        sql.param("password","root");
        sql.connect();
    }
    ~test()
    {
    }
    bool load_data(data::page &c,int id)
    {
        sql<<   "SELECT     title,content "
                "FROM       posts "
                "WHERE      id=?",id;
        dbixx::row r;
        if(!sql.single(r))
            return false;
        
        r >> c.title >> c.content;
        c.content = to_markdown(c.content);


        sql<<   "SELECT     content "
                "FROM       comments "
                "WHERE      post_id = ? "
                "ORDER BY   id ASC",id;

        dbixx::result res;
        sql.fetch(res);
        while(res.next(r)) {
            std::string comment;
            r >> comment;
            c.comments.push_back(to_markdown(comment));
        }
        return true;
    }
    
    void main(std::string /*unused*/)
    {
        cppcms::http::request::form_type::const_iterator id_ptr;
        id_ptr = request().get().find("id");
        if(id_ptr == request().get().end()) {
            response().make_error_response(404);
            return ;
        }
        std::string id = id_ptr->second;

        if(cache().fetch_page(id))
            return;

        data::page c;
        if(!load_data(c,atoi(id.c_str()))) {
            response().make_error_response(404);
            return;
        }
        render("page",c);
        cache().store_page(id,15);
    }
private:
    dbixx::session sql; 
};

int main(int argc,char ** argv)
{
    try {
        cppcms::service app(argc,argv);
        app.applications_pool().mount(cppcms::applications_factory<test>());
        app.run();
    }
    catch(std::exception const &e) {
        std::cerr<<e.what()<<std::endl;
    }
}
// vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
